-------------------------------------------------------------------------------
--
-- Title       : instr_controller
-- Design      : sw_backen
-- Author      : Vegard
-- Company     : NTNU
-- Description : Controller for handling instructions. 
-------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity instr_controller is						  
	generic(			
		-- Instruction size, modification requires redefining instructions
		INSTR_SIZE : natural :=16
	);
	port(
		clk				: in std_logic;
		instruction 	: in std_logic_vector(INSTR_SIZE-1 downto 0);
		read_ack 		: in std_logic;	
		write_ack		: in std_logic;
		hw_complete		: in std_logic;  
		cs				: in std_logic_vector(0 to 0);
		address			: out std_logic_vector(7 downto 0);
		shift			: out std_logic;
		register_write	: out std_logic;
		hw_en			: out std_logic;
		hw_we			: out std_logic;
		shift_mask		: out std_logic;
		zero_mask		: out std_logic;
		hw_run			: out std_logic;
		instr_complete  : out std_logic
	    );

end instr_controller;


architecture RTL of instr_controller is

type state is (IDLE,EXECUTING,WAITING);
signal cstate,nstate : STATE;	

signal shift_i  : std_logic; 
signal shift_ii : std_logic;
signal count 	: std_logic_vector(1 downto 0);
signal shift_complete : std_logic;
signal instr_complete_i : std_logic;

signal hw_en_i, hw_we_i : std_logic;
signal hw_run_i : std_logic;
signal sync, interrupt : std_logic;

begin	 
	
	
	NSTATE_PROC : process(sync,interrupt,cstate,instr_complete_i) is
	begin
		case cstate is 
			when IDLE =>	 
				if sync='1' then 
					nstate <= EXECUTING;
				else
					nstate <= IDLE;
				end if;
			when EXECUTING	=>
				if instr_complete_i ='1' then
					nstate <= WAITING;
				else
					nstate <= EXECUTING;
				end if;	  
			when WAITING =>		
				if interrupt ='1' then
					nstate <= IDLE;	 
				else
					nstate <= WAITING;
				end if;
		end case;
	end process NSTATE_PROC;
	
	CSTATE_PROC : process(clk) is
	begin
		if clk'event and clk='1' then
			cstate <= nstate;	
		end if;
	end process CSTATE_PROC;
	
	
	SHIFT_PROC : process(cstate,shift_i,shift_complete) is
	begin							   
		if cstate = WAITING then
			shift_ii <= '0';
		else
			shift_ii <= shift_i;
		end if;
	end process SHIFT_PROC;
					
	INSTR_COMPLETE_PROC : process(cstate,instr_complete_i) is
	begin
		if cstate = WAITING then
			instr_complete <= '1';
		else
			instr_complete <= instr_complete_i;
		end if;
	end process INSTR_COMPLETE_PROC;
	
	ACK_PROC : process(cstate,hw_en_i,hw_we_i) is
	begin
		if cstate = EXECUTING then
			hw_en <= hw_en_i;
			hw_we <= hw_we_i;
		else
			hw_en <= '0';
			hw_we <= '0';
		end if;
	end process;
	
	HW_RUN_PROC : process(hw_run_i,cstate) is
	begin
		if cstate =EXECUTING then 
			hw_run <= hw_run_i;
		else
			hw_run <='0';
		end if;
	end process HW_RUN_PROC;
	
	
	INSTRUCTION_DECODER : entity instr_decoder
		generic map(
			INSTR_SIZE => INSTR_SIZE
		)
		port map(
			opcode => instruction(INSTR_SIZE-1 downto INSTR_SIZE-2),
			flags =>  instruction(INSTR_SIZE-3 downto INSTR_SIZE-4),
			operand => instruction(INSTR_SIZE-5 downto 0),
			cs => cs,
			sync => sync,
			interrupt => interrupt,
			read_ack => read_ack,
			write_ack => write_ack,
			shift_complete => shift_complete,
			hw_complete => hw_complete,
			address => address,
			register_write => register_write,
			hw_en => hw_en_i,
			hw_we => hw_we_i, 
			count => count,
			shift_mask => shift_mask,
			zero_mask => zero_mask,
			hw_run  => hw_run_i,
			shift_o => shift_i,
			instr_complete => instr_complete_i
		);
	
		SHIFT_CONTROLLER : entity shift_alt
		generic map(C_COUNT_SIZE => 2)
		port map(
			clk 			=> clk,	 
			shift_i			=> shift_ii,
			count			=> count,
			shift_o			=> shift,
			shift_complete	=> shift_complete
		); 
	


end RTL;
