-------------------------------------------------------------------------------
--
-- Title       : RAM_master
-- Design      : ram_migration
-- Author      : Vegard
-- Company     : NTNU
-- File        : RAM_master.vhd
-- Description : Master for requesting operations on RAM migrated across reconfigurable
--			     interface
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;


entity RAM_master is
	generic(	 
		DATA_SIZE			: natural   :=32;
		SHIFT_SIZE			: natural   :=8;	   
		ADDR_SIZE			: natural   :=8
	);
	port(
		clk		: in std_logic;				   
		start	: in std_logic;				  
		shift	: in std_logic;
		grant	: in std_logic;
		mode	: in std_logic_vector(1 downto 0); 	
		data_i 	: in std_logic_vector(DATA_SIZE-1 downto 0);
		shift_i : in std_logic_vector(SHIFT_SIZE-1 downto 0);
		addr 	: in std_logic_vector(ADDR_SIZE-1 downto 0);
		request : out std_logic_VECTOR(1 downto 0);
		data_o	: out std_logic_vector(DATA_SIZE-1 downto 0);
		shift_o : out std_logic_vector(SHIFT_SIZE-1 downto 0);
		complete: out std_logic
	);
end RAM_master;



architecture RAM_master of RAM_master is

constant OP_IDLE  : std_logic_vector(1 downto 0) :="00";
constant OP_READ  : std_logic_vector(1 downto 0) :="01";
constant OP_WRITE : std_logic_vector(1 downto 0) :="10";
constant OP_RESET : std_logic_vector(1 downto 0) :="11";

type  STATE is (IDLE,ADDR_REQ,OP_REQ);

signal cstate,nstate 	: STATE;
signal write 		: std_logic;

signal reg0_i: std_logic_vector(DATA_SIZE-1 downto 0);

begin
	
	--Data transfer register
	REG0 : entity register_macro
		generic map(
			C_SIZE 		=> DATA_SIZE/4,
			C_NUM_REG  => 4
		)
		port map(
			clk =>clk,
			shift => shift,
			write => write,
			shift_i => shift_i,
			data_i => reg0_i,
			data_o => data_o,
			shift_o	=> shift_o
		);
		
		
		
		
	MODE_REQ_PROC : process(cstate) is
	begin 
		case cstate is 			 
			when ADDR_REQ =>
				 request <= OP_WRITE;
			when OP_REQ =>
				request <= mode; 
			when others =>
				request <= OP_IDLE;
		end case;
				
	end process MODE_REQ_PROC;
	
	WRITE_PROC : process(cstate,grant) is
	begin
		if cstate = IDLE and start ='1'then
			write <='1';	
		elsif cstate = ADDR_REQ and grant ='1' then
			write <= '1';
		else
			write <='0';
		end if;
		
	end process WRITE_PROC;
	
	COMPLETE_PROC : process(cstate,grant) is
	begin
		if cstate = OP_REQ and grant ='1' then
			complete <='1';
		else
			complete <='0';
		end if;
	end process COMPLETE_PROC;
	
	REG0_IN_PROC : process(cstate,addr,data_i) is
	begin
		if cstate = IDLE then
			reg0_i(ADDR_SIZE-1 downto 0) <= addr;
			reg0_i(DATA_SIZE-1 downto ADDR_SIZE) <= (others =>'0');
		else
			reg0_i <= data_i;
		end if;
	end process REG0_IN_PROC;
	
		
	NSTATE_PROC : process(cstate,start,grant) is
	begin
		case cstate is
			when IDLE =>
				if start ='1' then	   
					nstate <= ADDR_REQ;
				else
					nstate <= IDLE;
				end if;			   
				
			when ADDR_REQ =>			   
				if grant ='1' then
					nstate <= OP_REQ;
				else
					nstate <= ADDR_REQ;
				end if;					
			when OP_REQ =>
				if grant ='1' then
					nstate <=IDLE;
				else
					nstate <= OP_REQ;
				end if;		 
				
			when others => 
				null;
		end case;
	end process NSTATE_PROC;
	
	CSTATE_PROC : process(clk) is
	begin
		if clk'event and clk='1' then
			cstate <= nstate;
		end if;
	end process CSTATE_PROC;
		
end RAM_master;
