-------------------------------------------------------------------------------
--
-- Title       : RAM_slave
-- Design      : ram_migration
-- Author      : Vegard
-- Company     : NTNU
-- File        : RAM_slave.vhd
-- Description : 
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_unsigned.all;

entity RAM_slave is
	generic(	 
		DATA_SIZE			: natural   :=32;
		SHIFT_SIZE			: natural   :=8;	   
		ADDR_SIZE			: natural   :=8
	);
	port(
		clk		: in std_logic;				   
		reset	: in std_logic;
		request : in std_logic_VECTOR(1 downto 0);
		shift_i : in std_logic_vector(SHIFT_SIZE-1 downto 0);
		data_i	: in std_logic_vector(DATA_SIZE-1 downto 0);
		addr	: out std_logic_vector(ADDR_SIZE-1 downto 0);
		shift_g	: out std_logic;
		grant	: out std_logic;
		ram_rst : out std_logic;
		ram_wr	: out std_logic_vector(0 downto 0);
		ram_rd  : out std_logic
	);
end RAM_slave;



architecture RTL of RAM_slave is


constant OP_IDLE  : std_logic_vector(1 downto 0) :="00";
constant OP_READ  : std_logic_vector(1 downto 0) :="01";
constant OP_WRITE : std_logic_vector(1 downto 0) :="10";
constant OP_RESET : std_logic_vector(1 downto 0) :="11";


type STATE is (IDLE,ADDR_SHIFT,OP_BRANCH,READING,SHIFTING,WRITING,RESETTING);
signal cstate,nstate : STATE;

signal request_buf 	: std_logic_vector(1 downto 0);
signal count	   	: std_logic_vector(1 downto 0);

signal addr_buf		: std_logic_vector(ADDR_SIZE-1 downto 0);

begin
	
	addr <= addr_buf;
	
	NSTATE_PROC : process(cstate,request,request_buf,count) is
	begin
		case cstate is
			when IDLE =>   
				--signaling address transfer
				if request = OP_WRITE then	   
					nstate <= ADDR_SHIFT;
				elsif request = OP_RESET then
					nstate <= RESETTING;
				else
					nstate <= IDLE;
				end if;			   
				
			when ADDR_SHIFT =>
			    -- Assuming 8 bit address (1 shift operation on 8 bit wide bus)
				nstate <= OP_BRANCH;
			
			when OP_BRANCH =>
				if request = OP_READ then
					nstate <= READING;
				elsif request = OP_WRITE then 
					nstate <= SHIFTING;
				else
					nstate <= RESETTING;
				end if;	 
			when READING =>
			    --wait for BRAM read-ack then
				nstate <= SHIFTING;
				
			when SHIFTING =>	   
				-- shift four times
				if count ="11" then
					if request_buf = OP_READ then 
						nstate <= IDLE;
					else
						nstate <= WRITING;
			    	end if;
				end if;			   		 
			when WRITING =>
				-- write content to BRAM
				nstate <= IDLE;
			when RESETTING =>
				-- reset BRAM
		    	nstate <= IDLE;
			
				when others => 
				null;
		end case;
	end process NSTATE_PROC;
	
	CSTATE_PROC : process(clk) is
	begin
		if clk'event and clk='1' then
			cstate <= nstate;
		end if;
	end process CSTATE_PROC;

	REQ_BUF_PROC : process(clk) is
	begin
		if clk'event and clk ='1' then
			if cstate = OP_BRANCH then
				request_buf <= request;
			else
				request_buf <= request_buf;
			end if;	 
		end if;
	end process REQ_BUF_PROC;
	
	ADDR_BUF_PROC : process(clk) is
	begin
		if clk'event and clk='1' then
			if cstate = OP_BRANCH then
				addr_buf <= data_i(DATA_SIZE-1 downto DATA_SIZE-ADDR_SIZE);
			end if;
		end if;
	end process ADDR_BUF_PROC;
	
	COUNT_PROC : process(clk) is
	begin
		if clk'event and clk ='1' then
			if reset ='1' then
				count <= (others =>'0');
			elsif cstate = SHIFTING then
				count <= count+1;
			end if;
		end if;
	end process COUNT_PROC;
	
	SHIFT_PROC : process(cstate) is
	begin
		if cstate = ADDR_SHIFT then
			shift_g <='1';
		elsif cstate = SHIFTING then
			shift_g <='1';
		else
			shift_g <='0';
		end if;
	end process SHIFT_PROC;	  
	
	GRANT_PROC : process(cstate, request_buf) is
	begin
		if cstate = ADDR_SHIFT then 
			grant <='1';
		elsif request_buf = OP_WRITE and cstate = WRITING then
			grant <='1';
		elsif request_buf = OP_READ and cstate = READING then
			grant <='1';
		else
			grant <='0';
		end if;
	end process GRANT_PROC;
	
	RAM_RD_PROC : process(cstate) is
	begin
		if cstate = READING then
			ram_rd <='1';
		else
			ram_rd <='0';
		end if;
	end process RAM_RD_PROC;

	RAM_WR_PROC : process(cstate) is
	begin	   
		if cstate = WRITING then
			ram_wr <="1";
		else
			ram_wr <="0";
		end if;
	end process RAM_WR_PROC;
	
	RAM_RST_PROC : process(cstate) is
	begin
		if cstate = RESETTING then
			ram_rst <='1';
		else
			ram_rst <='0';	  
		end if;
	end process RAM_RST_PROC;
	
	
	
end RTL;
