-------------------------------------------------------------------------------
--
-- Title       : sequential_multiplier
-- Design      : reconfiguration
-- Author      : Vegard Endresen
-- Company     : NTNU
-- Description : Sequential multiplier using serial shift registers
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_unsigned.all;


entity sequential_multiplier is
generic(
		C_REG_SIZE    		: natural	:=32;
		C_BIT_WIDTH			: natural   :=8;
		C_LOG_OPERAND_SIZE	: natural   :=5;
		C_ADDR_SIZE			: natural   :=4 
		);

	port(
		clk			: in  std_logic;
		run			: in  std_logic;
		write		: in  std_logic;
		addr		: in  std_logic_vector(C_ADDR_SIZE-1 downto 0);
		data_i		: in  std_logic_vector(C_BIT_WIDTH-1 downto 0);
		data_o		: out std_logic_vector(C_BIT_WIDTH-1 downto 0);
		HW_complete	: out std_logic 
	);

end sequential_multiplier;


architecture RTL of sequential_multiplier is

--STATE FIELDS
----------------------------------------------------------------------
signal cstate,nstate : std_logic_vector(1 downto 0);

constant IDLE : std_logic_vector(1 downto 0) :="00";
constant LOAD : std_logic_vector(1 downto 0)	:="01";
constant SHIFTING : std_logic_vector(1 downto 0) :="10";
constant COMPLETE : std_logic_vector(1 downto 0) :="11";
----------------------------------------------------------------------

	   
signal cnt  	 : std_logic_vector(C_LOG_OPERAND_SIZE-1 downto 0);


signal sel : std_logic_vector(16 downto 0);
signal global_writes : std_logic_vector(16 downto 0);



--OPERAND A FIELDS
----------------------------------------------------------------------
signal OPERAND_A_REG_shift : std_logic;
signal OPERAND_A_REG_data_o : std_logic_vector(C_REG_SIZE-1 downto 0);
signal shift_isig : std_logic_vector(C_REG_SIZE/C_BIT_WIDTH-1 downto 0);
signal shift_osig : std_logic_vector(C_REG_SIZE/C_BIT_WIDTH-1 downto 0);
----------------------------------------------------------------------

--OPERAND B FIELDS
----------------------------------------------------------------------
signal OPERAND_B_REG_data_o : std_logic_vector(C_REG_SIZE-1 downto 0);
signal addend_b 			: std_logic_vector(C_REG_SIZE-1 downto 0);
----------------------------------------------------------------------

--RESULT_REG FIELDS
----------------------------------------------------------------------
signal RESULT_REG_shift 	: std_logic; 
signal RESULT_REG_write_l 	: std_logic; 
signal RESULT_REG_write 	: std_logic; 
signal partial_sum		 : std_logic_vector(C_REG_SIZE  downto 0);
signal partial_operand	 : std_logic_vector(2*C_REG_SIZE-1  downto 0); 
----------------------------------------------------------------------

signal STATE_REG_write : std_logic;
signal CNT_REG_write, CNT_REG_count : std_logic;


signal outputs : std_logic_vector(4*C_REG_SIZE+ C_BIT_WIDTH-1 downto 0);




begin
	
		outputs <=  '0' & cnt & cstate & partial_operand & OPERAND_B_REG_data_o & OPERAND_A_REG_data_o;
	
		RESULT_REG_write <= RESULT_REG_write_l or write;
	

		DECODER : entity decoder_custom
		port map(
			addr => addr,
			output => sel
		);
		
		MUX : entity mux_custom
			port map( 
				input => outputs,
				addr  => addr,
				output => data_o
			);
	  
	
	
	OPERAND_A_GEN : for i in 0 to (C_REG_SIZE/C_BIT_WIDTH)-1 generate
		--OPERAND A
		OPERAND_A_REG : entity si_shift_register
		generic map(C_SIZE 	=> C_BIT_WIDTH)
		port map(
			clk 	=> clk,
			write   => global_writes(i),
			shift_i   => shift_isig(i),
			shift_o   => shift_osig(i),
			shift 	=> OPERAND_A_REG_shift,
			data_i  => data_i,
			data_o  => OPERAND_A_REG_data_o((i+1)*C_BIT_WIDTH-1 downto i*C_BIT_WIDTH)
		);
	
	end generate;
	
	
	OPERAND_B_GEN : for i in C_REG_SIZE/C_BIT_WIDTH to 2*C_REG_SIZE/C_BIT_WIDTH-1 generate
		--OPERAND B
		OPERAND_B_REG : entity si_register
		generic map(C_SIZE 	=> C_BIT_WIDTH)
		port map(
			 clk 	  => clk,
		 	 write	  => global_writes(i),
		 	 data_i   => data_i,
		 	 data_o  => OPERAND_B_REG_data_o((i+1-C_REG_SIZE/C_BIT_WIDTH)*C_BIT_WIDTH-1 downto (i-C_REG_SIZE/C_BIT_WIDTH)*C_BIT_WIDTH)
		);
	end generate;
	
	
	
	--PARTIAL PRODUCT / RESULT
	RESULT_REG : entity di_gshift
	generic map(
		C_SIZE  => 2*C_REG_SIZE,
		C_NUM_REGS => 2*C_REG_SIZE/C_BIT_WIDTH,  
		C_INPUT_SIZE => C_REG_SIZE+1
	)
	port map(
		clk 	=> clk,
		write0	=> RESULT_REG_write_l,
		write1	=> global_writes(4*C_REG_SIZE/C_BIT_WIDTH-1 downto 2*C_REG_SIZE/C_BIT_WIDTH),
		data_i0 => partial_sum,
		data_i1 => data_i,
		shift 	=> RESULT_REG_shift,
		data_o  => 	partial_operand
	);			 

	
	CNT_REG	: entity si_register_count 
		generic map(C_SIZE => C_LOG_OPERAND_SIZE)
		port map(
			clk => clk,
			count => CNT_REG_count,
			write => global_writes(16),
			data_i	=> data_i(C_LOG_OPERAND_SIZE+1 downto 2),
			data_o  => cnt
	) ;
	
	STATE_REG : entity di_register
	generic map(
		C_SIZE => 2
	)						  
	port map (
	  clk 	 => clk,
	  write0 => STATE_REG_write,
	  write1 =>  global_writes(16),
	  data_i0 => nstate,
	  data_i1 => data_i(1 downto 0),
	  data_o  => cstate
	);
	
	
	
	ADDER : entity cla_adder
	generic map(OPERAND_SIZE => C_REG_SIZE)
							
	port map(
		a_i => partial_operand(2*(C_REG_SIZE-1)downto C_REG_SIZE-1),
		b_i => addend_b,
		s_o => partial_sum
	);
	
	NSTATE_PROC : process(cstate, run,cnt)
	begin	
		if run ='0' then
			nstate <= IDLE; 
		else
			case cstate is
				when IDLE =>
					nstate <= LOAD;
				when LOAD =>
					nstate <= SHIFTING;
				when SHIFTING =>
					if cnt = C_REG_SIZE-1 then
						nstate <= COMPLETE;
					else 
						nstate <= LOAD;
					end if;	 
				when COMPLETE =>
					nstate <= IDLE;	
				when others =>
			end case;
		end if;
	end process NSTATE_PROC;

	
	CNT_PROC : process(cstate,run) is 
	begin	
		if cstate = LOAD and run ='1' then
			CNT_REG_count <='1';
		else
			CNT_REG_count <= '0';
		end if;	
		
	end process CNT_PROC;
	
	
	WRITE_PROC : process(run,cstate,cnt) is 
	begin
		if (cstate = LOAD or cstate = COMPLETE) and run='1' then
			RESULT_REG_write_l <='1';
		else
			RESULT_REG_write_l <='0';
		end if;
	end process WRITE_PROC;
	
	SHIFT_PROC : process(cstate,run) is
	begin
		if cstate = SHIFTING and run='1' then
			RESULT_REG_shift 	<='1';
			OPERAND_A_REG_shift <='1';
		else
			OPERAND_A_REG_shift <='0'; 
			RESULT_REG_shift 	<='0';
		end if;
	end process SHIFT_PROC;
	
	ADDEND_B_PROC : process(OPERAND_A_REG_data_o,OPERAND_B_REG_data_o)is
	begin
		if OPERAND_A_REG_data_o(0) ='1' then
			addend_b <= OPERAND_B_REG_data_o;
		else
			addend_b <= (others =>'0');
		end if;
			
		
	end process ADDEND_B_PROC;

	
	FINISH_PROC : process(clk)
	begin					
		if clk'event and clk='1' then
			if cstate = COMPLETE then
				HW_complete <='1';
			else
				HW_complete <='0';	  
			end if;		   
			
		end if;
		
	end process FINISH_PROC;
	
	SHIFT_CHAIN_PROC : process(shift_osig) is
	begin		 
		shift_isig(C_REG_SIZE/C_BIT_WIDTH-1) <='0';
		for i in  C_REG_SIZE/C_BIT_WIDTH-2 downto 0	loop
			shift_isig(i) <= shift_osig(i+1);
		end loop;
		
	end process SHIFT_CHAIN_PROC;
	
	STATE_REG_WRITE_PROC : process(run) is
	begin
		if run ='1' then
			STATE_REG_write <='1';
		else
			STATE_REG_write <='0';
		end if;
	end process STATE_REG_WRITE_PROC;		
			
	GLOBAL_WRITE_PROC : process(sel,write) is
	begin
		for i in 0 to 16 loop
			global_writes(i) <= sel(i) and write;
		end loop;
		
	end process GLOBAL_WRITE_PROC;
	
	
end RTL;
