-------------------------------------------------------------------------------
--
-- Title       : sequential_multiplier
-- Design      : serial_architecture
-- Author      : Vegard
-- Company     : NTNU
-- Description : Parameterized sequential multiplicator, computing A*B
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_unsigned.all;


entity sequential_multiplier is
generic(
		OPERAND_SIZE      	: natural	:=4;
		LOG_OPERAND_SIZE	: natural   :=2
		);

	port(
		shift_i 	: in 	std_logic;
		shift_g		: in 	std_logic;
		clk			: in 	std_logic;
		run			: in 	std_logic; 
		shift_o		: out 	std_logic;
		hw_complete	: out	std_logic 
	);

end sequential_multiplier;


architecture RTL of sequential_multiplier is

--STATE FIELDS
----------------------------------------------------------------------
signal cstate,nstate : std_logic_vector(1 downto 0);

constant IDLE : std_logic_vector(1 downto 0) 	 :="00";
constant LOAD : std_logic_vector(1 downto 0)	 :="01";
constant SHIFTING : std_logic_vector(1 downto 0) :="10";
constant COMPLETE : std_logic_vector(1 downto 0) :="11";   

signal STATE_REG_write	 : std_logic;
signal STATE_REG_shift_o : std_logic;
----------------------------------------------------------------------

--COUNTER FIELDS
----------------------------------------------------------------------	   
signal cnt  	 		: std_logic_vector(LOG_OPERAND_SIZE-1 downto 0);
signal CNT_REG_count 	: std_logic;
signal CNT_REG_shift	: std_logic;
signal CNT_REG_shift_o  : std_logic;
----------------------------------------------------------------------

--OPERAND_A FIELDS 
----------------------------------------------------------------------
signal OPERAND_A_REG_shift 	 : std_logic;
signal OPERAND_A_REG_shift_l : std_logic;
signal OPERAND_A_REG_shift_o : std_logic; 
----------------------------------------------------------------------

--OPERAND_B FIELDS 
----------------------------------------------------------------------
signal OPERAND_B_REG_par_o 		: std_logic_vector(OPERAND_SIZE-1 downto 0);
signal OPERAND_B_REG_shift_o 	: std_logic;
signal addend_b 				: std_logic_vector(OPERAND_SIZE-1 downto 0);
----------------------------------------------------------------------


--RESULT FIELDS 
----------------------------------------------------------------------
signal partial_sum		 : std_logic_vector(OPERAND_SIZE  downto 0);
signal partial_operand	 : std_logic_vector(2*OPERAND_SIZE-1  downto 0);
signal RESULT_REG_shift	 : std_logic;
signal RESULT_REG_shift_l: std_logic;
signal RESULT_REG_write  : std_logic; 
signal addend_a 		 : std_logic_vector(OPERAND_SIZE-1 downto 0);
----------------------------------------------------------------------

begin
	
-- STRUCTURAL
----------------------------------------------------------------------
	OPERAND_A_REG : entity shift_registerSI
	generic map(C_SIZE 			=> OPERAND_SIZE)
	port map(
		clk 	=> clk,
		shift 	=> OPERAND_A_REG_shift,
		shift_i => shift_i,
		shift_o => OPERAND_A_REG_shift_o
	);
	
	
	OPERAND_B_REG : entity shift_registerSI
	generic map(C_SIZE 			=> OPERAND_SIZE)
	port map(
		clk 	=> clk,
		shift 	=> shift_g ,
		shift_i => OPERAND_A_REG_shift_o,
		shift_o => OPERAND_B_REG_shift_o,
		par_o 	=> OPERAND_B_REG_par_o	
	);

	CNT_REG : entity shift_register_counter 
		generic map(C_SIZE => LOG_OPERAND_SIZE)
		port map(
			clk => clk,
			shift =>   shift_g,			
			count   => CNT_REG_count,
			shift_i => OPERAND_B_REG_shift_o,
			shift_o => CNT_REG_shift_o,
			par_o   => cnt
		);
		
	STATE_REG : entity shift_register
		generic map(C_SIZE => 2)
		port map(
			clk => clk,
			par_i => nstate,
			par_o => cstate,
			par_write => STATE_REG_write,
			shift => shift_g,
			shift_i => CNT_REG_shift_o,
			shift_o	=> STATE_REG_shift_o
		);
	--PARTIAL PRODUCT / RESULT
	RESULT_REG : entity shift_register_PIOSIZE
	generic map(
		C_SIZE 			=> 2*OPERAND_SIZE,
		C_INPUT_SIZE   	=> OPERAND_SIZE+1 
	)
	port map(
		clk 	=> clk,
		par_i 	=> partial_sum,
		par_o 	=> partial_operand,
		shift 	=> RESULT_REG_shift,
		shift_i => STATE_REG_shift_o,
		shift_o => shift_o,
		par_write => RESULT_REG_write
	);			 
	
	ADDER : entity cla_adder
	generic map(OPERAND_SIZE => OPERAND_SIZE)
							
	port map(
		a_i => addend_a,
		b_i => addend_b,
		s_o => partial_sum
	); 
----------------------------------------------------------------------

-- BREHAVIOURAL
----------------------------------------------------------------------	
	NSTATE_PROC : process(cstate, run,cnt)
	begin	
		if run ='0' then
			nstate <= IDLE;
			STATE_REG_write <='0';
		else	
			STATE_REG_write <='1';
			case cstate is
				when IDLE =>
					nstate <= LOAD;
				when LOAD =>
					nstate <= SHIFTING;
				when SHIFTING =>
					if cnt = OPERAND_SIZE-1 then
						nstate <= COMPLETE;
					else 
						nstate <= LOAD;
					end if;	 
				when COMPLETE =>
					nstate <= IDLE;	
				when others =>
			end case;
		end if;
	end process NSTATE_PROC;
	
	CNT_PROC : process(cstate) is 
	begin	
		if cstate = LOAD then
			CNT_REG_count <= '1';
		else
			CNT_REG_count <= '0';
		end if;		
	end process CNT_PROC;

	
	WRITE_PROC : process(cstate,cnt,run) is 
	begin
		if (cstate = LOAD or cstate = COMPLETE) and run='1' then
				RESULT_REG_write <='1';
			
		else
			RESULT_REG_write <='0';
		end if;
	end process WRITE_PROC;
	
	SHIFT_LOCAL_PROC : process(cstate,run,shift_g) is
	begin
		if run ='1' then
			if cstate = SHIFTING then
				RESULT_REG_shift_l 	<='1';
				OPERAND_A_REG_shift_l <='1';
			else
				OPERAND_A_REG_shift_l <='0'; 
				RESULT_REG_shift_l 	<='0';
			end if;		
		else
			RESULT_REG_shift_l 	<= shift_g;
			OPERAND_A_REG_shift_l <=shift_g;
		end if;
	end process SHIFT_LOCAL_PROC;		
	
	SHIFT_PROC : process(RESULT_REG_shift_l,OPERAND_A_REG_shift_l,shift_g)is
	begin
		RESULT_REG_shift <= RESULT_REG_shift_l or shift_g;
		OPERAND_A_REG_shift <= OPERAND_A_REG_shift_l or shift_g;  
	end process SHIFT_PROC;
	
	
	ADDENDS_PROC : process(OPERAND_A_REG_shift_o,OPERAND_B_REG_par_o,partial_operand)is
	begin  
		addend_a <=  partial_operand(2*(OPERAND_SIZE-1)downto OPERAND_SIZE-1);
		
		if OPERAND_A_REG_shift_o ='1' then
			addend_b <= OPERAND_B_REG_par_o;
		else
			addend_b <= (others =>'0');
		end if;
			
		
	end process ADDENDS_PROC;

	
	FINISH_PROC : process(clk,cstate)
	begin					 
		if clk ='1' and clk'event then
			if cstate = COMPLETE then
				HW_complete <='1';
			else
				HW_complete <='0';
			end if;
		end if;
		
	end process FINISH_PROC;
----------------------------------------------------------------------	
end RTL;
