-----------------------------------------------------------------------------------------------------
--
-- Title       : serial_controller
-- Design      : serial_controller
-- Author      : Vegard Endresen
-- Company     : NTNU
-- Description : Controller that synchronizes CPU and hardaware. 
--	
--  CPU signals:  
--
--	|CPU_run |CPU_shift|CPU_interrupt|Description 											   |
--  |--------|---------|-------------|---------------------------------------------------------|
--	|	1	 |	 0	   |	  0 	 | run hardware until an event occurs			 		   |
--  |--------|---------|-------------|---------------------------------------------------------|
--	|	0	 |	  1	   |	  1  	 | shift data according to operand and await next operation|
--  |--------|---------|-------------|---------------------------------------------------------|
--	|   0	 |	  0	   |	  0 	 | interrupt operation and jump to idle					   |
--  |--------|---------|-------------|---------------------------------------------------------|
--
-----------------------------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity serial_controller is
	generic(C_COUNT_SIZE : natural :=5);
	port(
		signal clk 					: in  std_logic;
		signal CPU_interrupt		: in  std_logic; 
		signal CPU_shift			: in  std_logic;
		signal CPU_run				: in  std_logic;
		signal HW_complete			: in  std_logic;
		signal count 				: in  std_logic_vector(C_COUNT_SIZE-1 downto 0);
		signal shift 				: out std_logic; 
		signal run					: out std_logic;
		signal controller_waiting	: out std_logic
	);
	
end serial_controller;


architecture serial_controller of serial_controller is

type STATE is (IDLE,SHIFTING,RUNNING,WAITING);

signal cstate,nstate : STATE;
signal start 			: std_logic;
signal shift_complete 	: std_logic; 

begin
	
	SHIFT_CONTROLLER: entity controller
	generic map(C_COUNT_SIZE => C_COUNT_SIZE)
	port map(
		clk => clk,	
		count => count,
    	start => start,
		shift => shift,
		shift_complete => shift_complete
	);
	
	
	NSTATE_PROC : process(cstate,CPU_interrupt,CPU_shift,CPU_run,shift_complete,HW_complete) is
	begin			   
		case cstate is 
			when IDLE =>						
				if CPU_shift ='1' then
						nstate <= SHIFTING;
				elsif CPU_run ='1' then 
						nstate <= RUNNING;
				else
					nstate <= IDLE;
			  	end if;
			when SHIFTING =>
				if CPU_interrupt ='1' then
					nstate <= IDLE;
				elsif shift_complete ='1' then
					nstate <= WAITING;
				else
					nstate <= SHIFTING;
				end if;
			when RUNNING =>			 
				if CPU_interrupt ='1' then
					nstate <=IDLE;
				elsif HW_complete='1' then
					nstate <= WAITING;
				else
					nstate <= RUNNING;	  
				end if;
			when WAITING =>	 
				if CPU_interrupt ='1' then
					nstate <= IDLE;
				else
					nstate <= WAITING;
				end if;
			end case;
	end process NSTATE_PROC;
	
	CSTATE_PROC : process(clk) is
	begin			   
		if(clk'event and clk='0') then
			cstate <= nstate;
		end if;
		
	end process CSTATE_PROC;
	
	SHIFT_PROC:  process (clk) is
	begin			
		if clk'event and clk ='1' then
			if CPU_shift ='1' and cstate = IDLE then
				start <= '1';
			else
				start <= '0';
			end if;	 
		end if;
	end process SHIFT_PROC;
	
	RUN_PROC : process(cstate) is
	begin
		if cstate = RUNNING then 
			run <='1';
		else run <='0';
		end if;
			
	end process RUN_PROC;
	
	WAIT_PROC : process(cstate) is
	begin
		if cstate = WAITING then
			controller_waiting <='1';
		else
			controller_waiting <='0';
		end if;
		
	end process WAIT_PROC;
	
end serial_controller;
