///////////////////////////////////////////////////////////////////////////
//File			: d_cache.c
//Description	: device driver for data cache backend
//Notes			: --
//Author		: Vegard Endresen		
///////////////////////////////////////////////////////////////////////////

#include <linux/init.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h> 	/* printk() */
#include <linux/slab.h> 	/* kmalloc() */
#include <linux/fs.h> 		/* everything... */
#include <linux/errno.h> 	/* error codes */
#include <linux/types.h> 	/* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h> 	/* O_ACCMODE */
#include <asm/system.h> 	/* cli(), *_flags */
#include <asm/uaccess.h> 	/* copy_from/to_user */
#include <asm/io.h>		   /* readl(),writel()*/

#define mem_base_addr 0x81000000
#define reg_base_addr 0x81000400
#define mem_size 1024 		//BRAM address space
#define reg_size 8 		//SW-registers address space

#define device_name "d_backend"
#define byte_shift 8

MODULE_LICENSE("GPL");

/*
Separate pointers for remapped addresses.
Makes it possible separate BRAM and SWREGS
into non-contiguious memory spaces in system assembly
*/
static u32 *mem_base;
static u32 *reg_base;

int	  backend_major = 128;
/*Shared buffer for backend data transfer*/
char *backend_buffer;


/*Declaration of device driver functions*/
int 	backend_open(struct inode *inode, struct file *filp);
int 	backend_release(struct inode *inode, struct file *filp);
ssize_t backend_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
ssize_t backend_write(struct file *filp,const char *buf, size_t count, loff_t *f_pos);
void 	backend_exit(void);
int 	backend_init(void);

/*Fops structure*/
struct file_operations backend_fops = {
  read 	 : backend_read,
  write	 : backend_write,
  open	 : backend_open,
  release: backend_release
};

/*Declare load and unload functions*/
module_init(backend_init);
module_exit(backend_exit);


/*Transform PowerPC endian to linear bit string and vice versa*/
static int transform(int word){
	int mask,part,res,i;
	int shifts[] = {3,1,-1,-3};
	mask =0x000000FF;
	res =0;
	for(i=0; i<4; i++){
		if(shifts[i]>0)	part = (word &mask ) << (shifts[i]*byte_shift);
		else part = (word &mask ) >> (-shifts[i]*byte_shift);
		if(i==3) part = part & 	0x000000FF;	
		res = res | part;		
		mask = (mask <<	byte_shift);
	}
	return res;
}


/*Integer and character array converter functions*/

static void intToChar(char* buf, int value, int pos){
	int i,mask,tmp;
	mask = 0x000000FF;	
	for(i=0; i< 4; i++){
		tmp = (value & mask) >>(8*i);
		if(i==3) tmp = tmp & 0x000000FF;		;		
		buf[i+pos] =tmp;				
		mask = (mask <<8);
	}

}

static int charToInt(char *buf, int pos){
	int res,i,mask;
	mask = 0x000000FF;	
	res =0;
	for(i=0; i< 4; i++){
		res = res | ((buf[pos+i]<< 8*i) & mask);
		mask = (mask <<8);	
	}
	return res;
}



int backend_init(void){
  int res;
  printk(KERN_DEBUG "Driver loaded... \n");
  
  /*Register device*/
  res = register_chrdev(backend_major,device_name, &backend_fops);
  if (res < 0) {
    printk(KERN_ERR "Cannot register device %s %d\n",device_name, backend_major);
    return res;
  }

  printk(KERN_DEBUG "Remapping address space... \n");

  mem_base = ioremap(mem_base_addr,mem_size);
  if (mem_base == NULL) {
    printk(KERN_ERR "BRAM map failed\n");
    return -EIO;
  }

  reg_base = ioremap(reg_base_addr,reg_size);
  if (reg_base == NULL) {
    printk(KERN_ERR "SWREG map failed \n");
    return -EIO;
  }

  printk(KERN_DEBUG "Driver loaded and configured\n");

  return 0;
}

/*UNUSED*/
int backend_open(struct inode *inode, struct file *filp) {

  /*Success*/
  return 0;
}


ssize_t backend_read(struct file *filp, char *buf,size_t count, loff_t *f_pos){ 
  u32 i,data,reg_offset;
  ssize_t retval=0;
 
  //Verify that offset is not to big
  if(*f_pos > mem_size+reg_size){	
	goto fail;
  }
  //If count+offset exeeds memory space adjust count
  if(*f_pos+count> mem_size+reg_size){
	count = (mem_size+reg_size) -*f_pos; 
  }
  
  //Allocate memory for backend data
  backend_buffer = kmalloc(sizeof(char)*count, GFP_KERNEL);
  
   if (!backend_buffer) { 
    retval = -ENOMEM;
    goto fail; 
  } 


  for(i=0;i<count;i+=4){
   
	//read from BRAM
	if(*f_pos+i<mem_size){
		data = transform(readl(mem_base+(i+*f_pos/4)));  //compensate by 4, why?
	}
	else{
		reg_offset = (i+*f_pos)%mem_size;		 
		data = transform(readl(reg_base+reg_offset/4)); 
	}
	//copy data to buffer
 	intToChar(backend_buffer,data,i);
  }

  //transfer data to user space
  if(copy_to_user(buf,backend_buffer,count)){
	retval = -EFAULT;  	
	goto out;
  }
  retval = *f_pos;
  *f_pos += count;
  //retval = count;
 
  out:
	kfree(backend_buffer);
  fail:	
	//up(&dev->sem)
 	return retval;
}

ssize_t backend_write( struct file *filp, const char *buf,size_t count, loff_t *f_pos) {
  u32 i,data,reg_offset;
  ssize_t retval=0;

  //Verify that offset is not to big
  if(*f_pos > mem_size+reg_size){	
	goto fail;
  }
  //If count+offset exeeds memory space adjust count
  if(*f_pos+count> mem_size+reg_size){
	count = (mem_size+reg_size) -*f_pos; 
  }
  
  //Allocate memory for backend data
  backend_buffer = kmalloc(sizeof(char)*count, GFP_KERNEL);
  
   if (!backend_buffer) { 
    retval = -ENOMEM;
    goto fail; 
  } 

   //Transfer data from user space
  if(copy_from_user(backend_buffer,buf,count)){
	retval = -EFAULT;  	
	goto out;
  }

  for(i=0;i<count;i+=4){
	data = charToInt(backend_buffer,i);
	data = transform(data);
	//Read from BRAM
	if(i+*f_pos <mem_size){
		writel(data,mem_base+(i+*f_pos)/4); //compensate by 4, why?
	}
	else{
		reg_offset = (i+*f_pos)%mem_size;
		writel(data,reg_base+reg_offset/4);
	}
  }

  *f_pos += count;
  retval = count;
  
  out:
	kfree(backend_buffer);
  fail:	
 	return retval;
}


/*UNUSED*/
int backend_release(struct inode *inode, struct file *filp) {
 
  /*Success*/
  return 0;
}
void backend_exit(void){
  int ret;
  iounmap(mem_base);
  iounmap(reg_base);

  printk(KERN_ALERT "unregister device \n");
  ret = unregister_chrdev(backend_major, device_name);
  if (ret < 0) printk(KERN_ALERT "unregister fail: %d\n", ret);

  printk("Driver unloaded\n");
}
